package cat.hack3.codingtests.marsrover.cartography;

import java.util.List;
import java.util.logging.Logger;

/**
 * Notice that the representation is the following:
 * Latitude is Y, starts from 1 on the top and goes down incrementally
 * Longitude is X, starts from 1 on the left and goes to right incrementally
 *     1    2    3
 * 1  1-1  1-2  1-3
 * 2  2-1  2-2  2-3
 * 3  3-1  3-2  3-3
 */
public class MarsMap {
    Logger logger = Logger.getLogger(this.getClass().getName());
    private static final int FIRST_POSITION_IN_MAP = 1;
    private static final int INCREMENT_UNIT = 1;

    private final int height;
    private final int width;
    private Coordinates currentPosition;
    private final List<Coordinates> obstaclesLocalizations;

    private final MapIncrementalPositionResolver positionResolver;

    @SafeVarargs
    public MarsMap(int height, int width, Coordinates startingCoordinates, List<Coordinates>... obstaclesLocalizations) {
        this.height = height;
        this.width = width;
        currentPosition = startingCoordinates;
        this.obstaclesLocalizations = obstaclesLocalizations.length > 0 ? obstaclesLocalizations[0] : List.of();
        positionResolver = new MapIncrementalPositionResolver(FIRST_POSITION_IN_MAP, INCREMENT_UNIT);
    }

    public Coordinates updatePositionTowards(Direction direction) {
        boolean updatedPosition = switch (direction) {
            case NORTH -> decrementLatitude();
            case SOUTH -> incrementLatitude();
            case WEST -> decrementLongitude();
            case EAST -> incrementLongitude();
        };
        if (updatedPosition)
            logger.info(String.format("Updated coordinates towards %s, now is %d-%d",
                    direction, currentPosition.latitude(), currentPosition.longitude()));

        return currentPosition;
    }

    private boolean incrementLatitude() {
        int latitude = currentPosition.latitude();
        int newLatitude = positionResolver.getIncrementedPosition(latitude, height);
        return setNewLatitudeIfNoObstacle(newLatitude);
    }

    private boolean decrementLatitude() {
        int latitude = currentPosition.latitude();
        int newLatitude = positionResolver.getDecrementedPosition(latitude, height);
        return setNewLatitudeIfNoObstacle(newLatitude);
    }

    private boolean incrementLongitude() {
        int longitude = currentPosition.longitude();
        int newLongitude = positionResolver.getIncrementedPosition(longitude, width);
        return setNewLongitudeIfNoObstacle(newLongitude);
    }

    private boolean decrementLongitude() {
        int longitude = currentPosition.longitude();
        int newLongitude = positionResolver.getDecrementedPosition(longitude, width);
        return setNewLongitudeIfNoObstacle(newLongitude);
    }

    private boolean setNewLatitudeIfNoObstacle(int newLatitude) {
        Coordinates newCoordinates = currentPosition.ofUpdatedLatitude(newLatitude);
        return setNewPositionIfNoObstacle(newCoordinates);
    }

    private boolean setNewLongitudeIfNoObstacle(int newLongitude) {
        Coordinates newCoordinates = currentPosition.ofUpdatedLongitude(newLongitude);
        return setNewPositionIfNoObstacle(newCoordinates);
    }

    private boolean setNewPositionIfNoObstacle(Coordinates newCoordinates) {
        if (willCollideWithObstacle(newCoordinates))
            return false;
        else {
            currentPosition = newCoordinates;
            return true;
        }
    }

    private boolean willCollideWithObstacle(Coordinates newCoordinates) {
        boolean willCollide = obstaclesLocalizations.contains(newCoordinates);
        if (willCollide) {
            var collisionReport = String.format("There is an obstacle at %s, therefore Rover will keep the same position", newCoordinates);
            logger.info(collisionReport);
        }
        return willCollide;
    }

    public Coordinates getCurrentPosition() {
        return currentPosition;
    }
}
